// globals
float angle = 160;

// text
color black =color(0, 0, 0), grey=180, red=color(181, 7, 26, 255), yellow=color(241, 215, 3);
PFont Gotham;

// Earth
float radiusInner = 200;
float radiusOuter = 500;
PShape earthGlobe;
PImage earthImage;

// SPL
float splLat = 47.6067;
float splLon = -122.3325;
PVector splVector = getCartesianVector(splLat, splLon, radiusInner);
PVector splRefVector = getCartesianVector(splLat, splLon, radiusOuter);
  
// Locations
Table locationsTable;
PVector[] locationVectors;
PVector[] locationRefVectors;
float[] locationHeights;
float[] rotationAngles;
PVector[] rotationAxis;
int locationsCount;
int c;

void setup() {
  size(1280, 900, P3D);
  earthImage = loadImage("earth.jpg");
  noStroke();
  earthGlobe = createShape(SPHERE, radiusInner);
  earthGlobe.setTexture(earthImage);
  
  Gotham = createFont("Gotham-Bold.otf", 32);

  // setup locations
  locationsTable = loadTable("geocoded_labelled_20.csv", "header");
  locationsCount = locationsTable.getRowCount();
  locationVectors = new PVector[locationsCount];
  locationRefVectors = new PVector[locationsCount];
  locationHeights = new float[locationsCount];
  rotationAngles = new float[locationsCount];
  rotationAxis = new PVector[locationsCount];
  PVector xAxis = new PVector (0, 1, 0);

  for (int i = 0; i < locationsCount; i++) {
    TableRow row = locationsTable.getRow(i);
    
    float lat = row.getFloat("latitude");
    float lon = row.getFloat("longitude");
    float numberOfCheckouts = row.getFloat("NumberOfCheckouts");
    
    locationVectors[i] = getCartesianVector(lat, lon, radiusInner);
    locationRefVectors[i] = getCartesianVector(lat, lon, radiusOuter);
    
    float maxHeight = pow(10, 10);
    locationHeights[i] = map(pow(11, numberOfCheckouts), 0, maxHeight, 10, 500);
    
    rotationAngles[i] = PVector.angleBetween(xAxis, locationVectors[i]);
    rotationAxis[i] = xAxis.cross(locationVectors[i]);  
  }
}

void doRotation(float speed) {
  rotateY(angle);
  angle += speed;
}

void drawEarth() {
  noFill();
  //stroke(100);
  //sphere(radiusInner);
  noStroke();
  shape(earthGlobe);
}

void drawSPL() {  
  pushMatrix();
  translate(splVector.x, splVector.y, splVector.z);
  fill(204, 102, 0);  // random red colour
  box(10);
  popMatrix();
}

void drawCities() {

  for (int i = 0; i < locationsCount; i++) {
    pushMatrix();
    fill(255);
    PVector vector = locationVectors[i];
    PVector rotationAxes = rotationAxis[i];
    translate(vector.x, vector.y, vector.z);
    rotate(rotationAngles[i], rotationAxes.x, rotationAxes.y, rotationAxes.z);  // TODO couldn't figure this out
    box(5, locationHeights[i], 5);
    //box(5, 5, 5);
    popMatrix();
  }
}

void drawBezierCurves() {
  noFill();
  stroke(255);
  for (int i = 0; i < locationsCount; i++) {
    PVector vector = locationVectors[i];
    PVector refVector = locationRefVectors[i];
    beginShape();
    vertex(vector.x, vector.y, vector.z);
    bezierVertex(
        refVector.x, refVector.y, refVector.z,
        splRefVector.x, splRefVector.y, splRefVector.z,
        splVector.x, splVector.y, splVector.z
    );
    endShape();
  }
}

void drawText() {
  textSize(53);
  textFont(Gotham);
  fill(255);
  textSize(20);
  textAlign(RIGHT, TOP);
  text("This 3D visualisation shows which locations are mentioned in SPL library books", mouseX, mouseY);
}


void draw() {
  background(51);
  translate(width * 0.5, height * 0.5);
  
  doRotation(0.005);
  drawEarth();
  drawSPL();
  drawCities();  
  drawBezierCurves();
  drawText();
    
}


// Converting to cartesian coordinates
PVector getCartesianVector(float lat, float lon, float radius) {
  float radLat = radians(lat);
  float radLon = radians(lon) + PI;
  float x = radius * cos(radLat) * cos(radLon);
  float y = radius * cos(radLat) * sin(radLon);
  float z = radius * sin(radLat);
  
  // FYI previous conversion system was fucked - because in processing, x/y axis are on screen, z-axis comes _out_ of screen :-/
  // vs every actual cartesian transform algorithm we find online: x/y horizontal, z vertical
  return new PVector(x, -z, -y);
}